
-- [Ice Project 2012, turtlesort.com]

-- Main scene of game. This is where maps are
-- rendered, collision detection happens, and
-- player control handling occurs.

local scene = {}

function scene.init()	

	-- Tile/event handling
	scene.event = require 'event-manager'

	-- Required for on-screen text
	scene.dialog = require 'lib/dialog'
	scene.dialog.font = game.font.get(16)
	
	-- All maps should be placed in this table
	scene.maps = {
		'maps/map-intro',
		
		'maps/corridor0',
		'maps/puzzle0',
		
		--
		'maps/corridor1',
		'maps/puzzle1',
		
		'maps/corridor2',
		'maps/puzzle2',

		'maps/corridor3',
		'maps/puzzle3',
		
		'maps/corridor4',
		'maps/puzzle4',
		--]]
		
		'maps/corridor1', -- An intentional repeat
		'maps/corridor5', -- The game can end here if player encounters the child
		'maps/corridor6',
		'maps/corridor-ending',
		'maps/map-ending'
	}
	scene.currentmapindex = 1
	scene.currentmap = require(scene.maps[scene.currentmapindex]) -- initial map load
	
	-- Preload the tileset image
	scene.tileset = scene.currentmap.tilesets[1]
	scene.tilesetimage = g.newImage(scene.tileset.image)
	scene.x = 0
	scene.y = 0
	scene.alpha = 255
	
	scene.player = {}
	
	-- Preload player sprite
	scene.player.sprite = require 'player-sprite'
	scene.player.direction = 'onground'
	scene.childofwinter = {}
	scene.childofwinter.yoffset = 0
	
	-- Special effects
	scene.flash = {}
	
	-- debug
	scene.targetx = 0
	scene.targety = 0

end

function scene.load()
	g.setBackgroundColor(51,51,51) -- matches the tileset's bg

	-- Load the current map
	scene.currentmap = require(scene.maps[scene.currentmapindex])
	scene.layer = scene.currentmap.layers[1]
	scene.layer.totaltiles = scene.layer.width*scene.layer.height
	scene.quadcache = {}
	scene.rockcache = {}
	scene.overcache = {}
	
	-- Special case: we want the rock with alpha background to be in cache no matter what
	scene.quadcache[38] = love.graphics.newQuad(
		6 * scene.tileset.tilewidth,
		4 * scene.tileset.tilewidth,
		scene.tileset.tilewidth,
		scene.tileset.tileheight,
		512,  -- po2
		512   -- po2
		-- scene.tileset.imagewidth,
		-- scene.tileset.imageheight
	)
	--]]
	
	-- Special case: Wall overlay that is drawn over player
	scene.quadcache[39] = love.graphics.newQuad(
		7 * scene.tileset.tilewidth,
		4 * scene.tileset.tilewidth,
		scene.tileset.tilewidth,
		scene.tileset.tileheight,
		512,  -- po2
		512   -- po2
		-- scene.tileset.imagewidth,
		-- scene.tileset.imageheight
	)
	--]]
	
	-- Build the quadcache
	local totaltilesinrow = scene.tileset.imagewidth/scene.tileset.tilewidth
	
	for x = 0, scene.layer.width - 1 do
		for y = 0, scene.layer.height - 1 do
		
			-- This is the tile's index on the map
			local mapindex = x + scene.layer.width*y + 1
			
			-- We use to find the tile's index on the tileset
			-- (tileset indexes start at 0)
			local tilesetindex = scene.layer.data[mapindex] - 1
			
			-- Special case: We replace normal rock tiles with their
			-- associated terrain-only tile, then overlay the rock-with-alpha-background
			-- on top during rendering so we can unalign the rocks from the grid as we
			-- please
			if tilesetindex == 46 then			-- rock on ice
				scene.rockcache[mapindex] = 46
				tilesetindex = 9				-- change to terrain tile
			elseif tilesetindex == 47 then		-- rock on land
				scene.rockcache[mapindex] = 47
				tilesetindex = 12				-- change to associated terrain tile
			end
			
			-- Special case: We remember the door tiles we encounter so we can draw
			-- an overlay later on the top of the door frame. This gives the effect of
			-- the player going underneath the door frame!
			if tilesetindex == 31 then			-- door tile
				local doorOverlay = {}
				doorOverlay.x = x*scene.tileset.tilewidth
				doorOverlay.y = (y-1)*scene.tileset.tilewidth
				table.insert(scene.overcache,doorOverlay)
			end
			
			-- Cache the neccessary quad for the tile if not already done
			if not scene.quadcache[tilesetindex] then
			
				-- Now we want 0-indexed tile coordinates on the tileset
				local tilemapx = tilesetindex % totaltilesinrow
				local tilemapy = math.floor(tilesetindex/totaltilesinrow)
				
				-- And finally the exact coordinates for the tile on tileset
				local pixelx = tilemapx * scene.tileset.tilewidth
				local pixely = tilemapy * scene.tileset.tileheight
			
				scene.quadcache[tilesetindex] = love.graphics.newQuad(
					pixelx,
					pixely,
					scene.tileset.tilewidth,
					scene.tileset.tileheight,
					512,  -- po2
					512   -- po2
					-- scene.tileset.imagewidth,
					-- scene.tileset.imageheight
				)
				
			end

		end
	end
	
	-- Player data
	scene.player.finishedmap = false
	scene.player.canbecontrolled = true
	
	-- Starting position for player
	scene.player.tilex = scene.currentmap.properties.startx
	scene.player.tiley = scene.currentmap.properties.starty
	
	-- Center camera on player right away
	scene.x = scene.tileset.tilewidth * (7 - scene.player.tilex) -- 8,6 in tile coordinates is center of screen
	scene.y = scene.tileset.tilewidth * (5 - scene.player.tiley)
	scene.targetx = scene.x
	scene.targety = scene.y
	
	scene.player.canbecontrolled = false
	
	scene.event.maptrigger()

end

function scene.draw()

	g.setColor(255,255,255,scene.alpha)

	-- Draw the map (tiles under player)
	for x = 0, scene.layer.width - 1 do
		for y = 0, scene.layer.height - 1 do
		
			-- This is the tile's index on the map
			local mapindex = x + scene.layer.width*y + 1
			
			-- We use to find the tile's index on the tileset
			local tilesetindex = scene.layer.data[mapindex] - 1
			
			if tilesetindex == 46 then			-- rock on ice
				tilesetindex = 9				-- change to terrain tile
			elseif tilesetindex == 47 then		-- rock on land
				tilesetindex = 12				-- change to associated terrain tile
			end
			
			g.drawq(
				scene.tilesetimage,
				scene.quadcache[tilesetindex],
				
				-- Flooring the scene coordinates prevents tiles
				-- prevents individual tiles from slightly unaligning with
				-- the map during player movement
				math.floor(scene.x) + x*scene.tileset.tilewidth,
				math.floor(scene.y) + y*scene.tileset.tilewidth
			)
			
			-- Rocks with transparent backgrounds
			if scene.rockcache[mapindex] then
				g.drawq(
					scene.tilesetimage,
					scene.quadcache[38], -- a rock with transparent background
					math.floor(scene.x) + x*scene.tileset.tilewidth,
					math.floor(scene.y) + y*scene.tileset.tilewidth - 12
				)
			end
			--]]
			
		end
	end
	
	g.setColor(255,255,255,255)
	
	-- Character sprite
	scene.player.sprite.draw(
		scene.tileset.tilewidth * 7,
		scene.tileset.tilewidth * 5,
		scene.player.direction,
		scene.player.tilex,
		scene.player.tiley,
		scene.targetx == scene.x and scene.targety == scene.y
	)
	
	-- Child of Winter sprite (only appears after last puzzle)
	if scene.childofwinter.visible then
		g.drawq(
			scene.player.sprite.image,
			scene.player.sprite.quadcache['childofwinter'],
			scene.tileset.tilewidth * 7,
			math.floor(scene.y) + (9 * scene.tileset.tilewidth) - 12 + scene.childofwinter.yoffset
		)	
	end
	
	-- Draw the map tiles that go over the player (only the door frame tile
	-- goes over player right now)
	for i = 1, #scene.overcache do
		g.drawq(
			scene.tilesetimage,
			scene.quadcache[39],
			math.floor(scene.x) + scene.overcache[i].x,
			math.floor(scene.y) + scene.overcache[i].y
		)
	end
	
	-- Stage number top right
	if scene.currentmapindex > 1 then
		g.setColor(200,200,200)
		g.setFont(game.font.get(15))
		g.print('Stage 1-' .. math.ceil((scene.currentmapindex - 1)/2) ,g.getWidth() - 150,20)
	end
	
	-- Falling snow flakes
	game.snow.draw()
	
	-- Flash effect for encounter with the child of winter
	if scene.flash.on then
		if scene.flash.alpha > 255 then
			g.setColor(255, 255, 255, 255)
		else
			g.setColor(255, 255, 255, scene.flash.alpha)
		end
		g.rectangle('fill', 0, 0, g.getWidth(), g.getHeight())
	end
	
	-- On-screen game text
	scene.dialog.draw()
	
	-- DEBUG:
	-- g.setColor(255,0,0)
	-- g.print(scene.alpha, 250,250)
	-- g.print(scene.player.tilex .. ',' .. scene.player.tiley, 10, 10)
	-- local index = (scene.player.tilex+16*scene.player.tiley) + 1
	-- g.print(index ..': '.. scene.currentmap[index], 10, 20)
	-- g.print('target: ' .. scene.targetx .. ',' .. scene.targety,10,10)
	-- g.print('actual: ' .. scene.x .. ',' .. scene.y,10,25)
end

function scene.update(dt)

	-- The targetx/scene.targety are the pixel coordinates of where the
	-- map image should be. The target coordinates are calculated so
	-- that the player is always at the center of screen.
	scene.targetx = scene.tileset.tilewidth * (7 - scene.player.tilex) -- 8,6 in tile coordinates is center of screen
	scene.targety = scene.tileset.tilewidth * (5 - scene.player.tiley)
	local change = dt * scene.tileset.tilewidth * 5 -- the higher the number the faster the player moves

	-- scene.x/scene.y are the actual pixel coordinates of where
	-- the map image is. They drift towards scene.targetx/scene.targety over
	-- time making the "top-down camera" smoothly move from one tile
	-- to another.
	if scene.x < scene.targetx then
		scene.x = scene.x + change
	elseif scene.x > scene.targetx then
		scene.x = scene.x - change
	end
	
	if scene.y < scene.targety then
		scene.y = scene.y + change
	elseif scene.y > scene.targety then
		scene.y = scene.y - change
	end

	if scene.alpha ~= 255 then
		scene.alpha = scene.alpha + 200 * dt
		
		if scene.alpha > 255 then
			scene.alpha = 255
		end
	end
	
	-- Snap scene.x/scene.y to scene.targetx/scene.targety once they pass their target.
	-- Note that as the player moves right, the scene coordinates move left
	-- which is why the comparisons look backwards.
	if (scene.player.direction == 'right' and scene.x < scene.targetx) or
	   (scene.player.direction == 'left' and scene.x > scene.targetx) then
		scene.x = scene.targetx
	end
	if (scene.player.direction == 'up' and scene.y > scene.targety) or
	   (scene.player.direction == 'down' and scene.y < scene.targety) then
		scene.y = scene.targety
	end
	
	-- Process keyboard input only when the player isn't in the middle
	-- of moving from one tile to another (buttons that open up the
	-- menu can still be used and are handled in scene.keypressed())
	if scene.x == scene.targetx 
	and scene.y == scene.targety 
	and scene.player.canbecontrolled then
	
		local newCoordinateX = scene.player.tilex
		local newCoordinateY = scene.player.tiley
		local newDirection = scene.player.direction

		if love.keyboard.isDown('w') or love.keyboard.isDown('up') then
			newCoordinateY = scene.player.tiley - 1
			newDirection = 'up'
		elseif love.keyboard.isDown('s') or love.keyboard.isDown('down') then
			newCoordinateY = scene.player.tiley + 1
			newDirection = 'down'
		elseif love.keyboard.isDown('a') or love.keyboard.isDown('left') then
			newCoordinateX = scene.player.tilex - 1
			newDirection = 'left'
		elseif love.keyboard.isDown('d') or love.keyboard.isDown('right') then
			newCoordinateX = scene.player.tilex + 1
			newDirection = 'right'
		end
		
		scene.player.direction = newDirection
		
		-- Are we about to walk into something solid? The solid
		-- tiles of the map are all on the bottom half of the tileset.
		-- This makes all solid tiles have an index greater than 32 (the
		-- tileset has 64 tiles).
		local index = newCoordinateX + scene.layer.width*newCoordinateY + 1
		
		if index >= 0 and index <= scene.layer.totaltiles then

			if scene.layer.data[index] <= 32 then
							
				scene.player.tilex = newCoordinateX
				scene.player.tiley = newCoordinateY

				-- Did we just land on ice? If so we continue moving the player
				-- in the same direction until they hit something solid.
				while scene.layer.data[index] == 10 do

					if scene.player.direction == 'left' then
						newCoordinateX = scene.player.tilex - 1
					elseif scene.player.direction == 'right' then
						newCoordinateX = scene.player.tilex + 1
					elseif scene.player.direction == 'up' then
						newCoordinateY = scene.player.tiley - 1
					elseif scene.player.direction == 'down' then
						newCoordinateY = scene.player.tiley + 1
					end

					index = newCoordinateX + scene.layer.width*newCoordinateY + 1
					if scene.layer.data[index] <= 32 then
						scene.player.tilex = newCoordinateX
						scene.player.tiley = newCoordinateY
					end
				end
				
			end
		end
	
	end

	-- Check if player landed on end tile
	if scene.x == scene.targetx and scene.y == scene.targety then
	
		local endx = scene.tileset.tilewidth * (7 - scene.currentmap.properties.endx) -- 8,6 in tile coordinates is center of screen
		local endy = scene.tileset.tilewidth * (5 - scene.currentmap.properties.endy)
		   
		-- Is the end tile at the center of the screen?
		if scene.x == endx and scene.y == endy and not scene.player.finishedmap then
			scene.player.finishedmap = true
			
			-- Are there any more maps?
			if scene.currentmapindex + 1 <= #scene.maps then

				-- Set the index for the next map we want to load
				scene.currentmapindex = scene.currentmapindex + 1
			
				-- We want a transition animation between maps
				if scene.currentmapindex == 2 then
					game.getScene('fader').setTarget('falling')
					game.setCurrentScene('fader')
				else
					local fader = game.getScene('fader')
					fader.fadeOutTime = 0.35
					fader.fadeInTime = 0.35
					
					fader.setTarget('main')
					
					--fader.setTarget('blank')
					game.setCurrentScene('fader')
				end
				
				-- scene-fader will automatically switch back to
				-- scene-main after it is done
			
				-- oldway:
				-- scene.load()
			else
				-- end game
			end
			
		end
	end
	
	
	scene.dialog.update(dt)
	game.snow.update(dt)

	-- This code block is used for white flashes on the screen (the
	-- ones you get when you meet the child of winter).
	if scene.flash.on then
		if scene.flash.direction == 'up' then
			scene.flash.alpha = scene.flash.alpha + dt*700
			
			-- hackity hack hack for sfx
			if scene.flash.alpha > 350 and scene.flash.alpha < 500 then 
				if scene.flash.sfx then game.audio.play(scene.flash.sfx) end
				scene.flash.alpha = 500
				
			-- The alpha value goes past 255 so that the duration the flash
			-- blocks out the entire screen is a little longer. In the draw
			-- function, all alpha values greater than 255 are reduced to 255.
			elseif scene.flash.alpha > 600 then
				scene.flash.alpha = 255
				scene.flash.direction = 'down'

				if scene.flash.notify then
					scene.flash.notify()
				end
				
			end
		end
		
		if scene.flash.direction == 'down' then
		
			scene.flash.alpha = scene.flash.alpha - dt*300
			
			if scene.flash.alpha < 0 then
				scene.flash.alpha = 0
				scene.flash.on = false
				
				if scene.flash.endevent then
					scene.flash.endevent()
				end
				
			end		
		end
	end
	
end

-- Only used to handle menu events
function scene.keypressed(key,unicode)

	scene.dialog.keypressed(key,unicode)

	if key == 'escape' then
		love.event.push('q')
	end

end

return scene